@@ -24,6 +24,7 @@
 #endif
 
 #include <gst/gst.h>
+#include <stdio.h>
 #include <string.h>
 #include "gstmms.h"
 
@@ -240,23 +241,22 @@ gst_mms_prepare_seek_segment (GstBaseSrc * src, GstEvent * event,
   GstSeekFlags flags;
   GstFormat seek_format;
   gdouble rate;
-  GstMMS *mmssrc = GST_MMS (src);
 
   gst_event_parse_seek (event, &rate, &seek_format, &flags,
       &cur_type, &cur, &stop_type, &stop);
 
   if (seek_format != GST_FORMAT_BYTES && seek_format != GST_FORMAT_TIME) {
-    GST_LOG_OBJECT (mmssrc, "Only byte or time seeking is supported");
+    GST_LOG_OBJECT (src, "Only byte or time seeking is supported");
     return FALSE;
   }
 
   if (stop_type != GST_SEEK_TYPE_NONE) {
-    GST_LOG_OBJECT (mmssrc, "Stop seeking not supported");
+    GST_LOG_OBJECT (src, "Stop seeking not supported");
     return FALSE;
   }
 
   if (cur_type != GST_SEEK_TYPE_NONE && cur_type != GST_SEEK_TYPE_SET) {
-    GST_LOG_OBJECT (mmssrc, "Only absolute seeking is supported");
+    GST_LOG_OBJECT (src, "Only absolute seeking is supported");
     return FALSE;
   }
 
@@ -288,8 +288,8 @@ gst_mms_do_seek (GstBaseSrc * src, GstSegment * segment)
       return FALSE;
     }
     start = mmsx_get_current_pos (mmssrc->connection);
-    GST_LOG_OBJECT (mmssrc, "sought to %f sec, offset after seek: %lld\n",
-        (double) segment->start / GST_SECOND, start);
+    GST_INFO_OBJECT (mmssrc, "sought to %" GST_TIME_FORMAT ", offset after "
+        "seek: %" G_GINT64_FORMAT, GST_TIME_ARGS (segment->start), start);
   } else if (segment->format == GST_FORMAT_BYTES) {
     start = mmsx_seek (NULL, mmssrc->connection, segment->start, SEEK_SET);
     /* mmsx_seek will close and reopen the connection when seeking with the
@@ -298,11 +298,11 @@ gst_mms_do_seek (GstBaseSrc * src, GstSegment * segment)
       GST_DEBUG_OBJECT (mmssrc, "connection broken during seek");
       return FALSE;
     }
-    GST_DEBUG_OBJECT (mmssrc, "sought to: %llu bytes, result: %lld",
-        segment->start, start);
+    GST_INFO_OBJECT (mmssrc, "sought to: %" G_GINT64_FORMAT " bytes, "
+        "result: %" G_GINT64_FORMAT, segment->start, start);
   } else {
-    GST_DEBUG_OBJECT (mmssrc, "unsupported seek segment format: %d",
-        (int) segment->format);
+    GST_DEBUG_OBJECT (mmssrc, "unsupported seek segment format: %s",
+        GST_STR_NULL (gst_format_get_name (segment->format)));
     return FALSE;
   }
   gst_segment_init (segment, GST_FORMAT_BYTES);
@@ -452,6 +452,14 @@ gst_mms_start (GstBaseSrc * bsrc)
         gst_message_new_element (GST_OBJECT_CAST (mms),
             gst_structure_new ("redirect", "new-location", G_TYPE_STRING, url,
                 NULL)));
+
+    /* post an error message as well, so that applications that don't handle
+     * redirect messages get to see a proper error message */
+    GST_ELEMENT_ERROR (mms, RESOURCE, OPEN_READ,
+        ("Could not connect to streaming server."),
+        ("A redirect message was posted on the bus and should have been "
+            "handled by the application."));
+
     return FALSE;
   }
 
