@@ -1,5 +1,5 @@
 /* GStreamer
- * Copyright (C) 2008 Jan Schmidt <thaytan@noraisin.net>
+ * Copyright (C) 2008-2009 Jan Schmidt <thaytan@noraisin.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -21,11 +21,13 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
 #include <string.h>
 
 #include <gst/gst.h>
 #include <gst/gst-i18n-plugin.h>
 #include <gst/interfaces/navigation.h>
+#include <gst/video/video.h>
 
 #include "resindvdsrc.h"
 
@@ -96,8 +98,8 @@ static GstFormat chapter_format;
 
 static void rsn_dvdsrc_register_extra (GType rsn_dvdsrc_type);
 
-GST_BOILERPLATE_FULL (resinDvdSrc, rsn_dvdsrc, RsnPushSrc,
-    RSN_TYPE_PUSH_SRC, rsn_dvdsrc_register_extra);
+GST_BOILERPLATE_FULL (resinDvdSrc, rsn_dvdsrc, GstBaseSrc,
+    GST_TYPE_BASE_SRC, rsn_dvdsrc_register_extra);
 
 static gboolean read_vts_info (resinDvdSrc * src);
 
@@ -108,15 +110,15 @@ static void rsn_dvdsrc_get_property (GObject * object, guint prop_id,
 
 static void rsn_dvdsrc_finalize (GObject * object);
 
-static gboolean rsn_dvdsrc_start (RsnBaseSrc * bsrc);
-static gboolean rsn_dvdsrc_stop (RsnBaseSrc * bsrc);
-static gboolean rsn_dvdsrc_unlock (RsnBaseSrc * bsrc);
-static gboolean rsn_dvdsrc_unlock_stop (RsnBaseSrc * bsrc);
+static gboolean rsn_dvdsrc_start (GstBaseSrc * bsrc);
+static gboolean rsn_dvdsrc_stop (GstBaseSrc * bsrc);
+static gboolean rsn_dvdsrc_unlock (GstBaseSrc * bsrc);
+static gboolean rsn_dvdsrc_unlock_stop (GstBaseSrc * bsrc);
 
-static gboolean rsn_dvdsrc_is_seekable (RsnBaseSrc * bsrc);
-static gboolean rsn_dvdsrc_prepare_seek (RsnBaseSrc * bsrc, GstEvent * event,
+static gboolean rsn_dvdsrc_is_seekable (GstBaseSrc * bsrc);
+static gboolean rsn_dvdsrc_prepare_seek (GstBaseSrc * bsrc, GstEvent * event,
     GstSegment * segment);
-static gboolean rsn_dvdsrc_do_seek (RsnBaseSrc * bsrc, GstSegment * segment);
+static gboolean rsn_dvdsrc_do_seek (GstBaseSrc * bsrc, GstSegment * segment);
 static GstStateChangeReturn
 rsn_dvdsrc_change_state (GstElement * element, GstStateChange transition);
 
@@ -138,9 +140,11 @@ static void rsn_dvdsrc_check_nav_blocks (resinDvdSrc * src);
 static void rsn_dvdsrc_schedule_nav_cb (resinDvdSrc * src,
     RsnDvdPendingNav * next_nav);
 
-static GstFlowReturn rsn_dvdsrc_create (RsnPushSrc * psrc, GstBuffer ** buf);
-static gboolean rsn_dvdsrc_src_event (RsnBaseSrc * basesrc, GstEvent * event);
-static gboolean rsn_dvdsrc_src_query (RsnBaseSrc * basesrc, GstQuery * query);
+static gboolean rsn_dvdsrc_check_get_range (GstBaseSrc * src);
+static GstFlowReturn rsn_dvdsrc_create (GstBaseSrc * bsrc, guint64 offset,
+    guint length, GstBuffer ** buf);
+static gboolean rsn_dvdsrc_src_event (GstBaseSrc * basesrc, GstEvent * event);
+static gboolean rsn_dvdsrc_src_query (GstBaseSrc * basesrc, GstQuery * query);
 
 static GstClockTime ifotime_to_gsttime (dvd_time_t * ifo_time);
 static void rsn_dvdsrc_send_commands_changed (resinDvdSrc * src);
@@ -203,13 +207,11 @@ rsn_dvdsrc_class_init (resinDvdSrcClass * klass)
 {
   GObjectClass *gobject_class;
   GstElementClass *gstelement_class;
-  RsnBaseSrcClass *gstbasesrc_class;
-  RsnPushSrcClass *gstpush_src_class;
+  GstBaseSrcClass *gstbasesrc_class;
 
   gobject_class = (GObjectClass *) klass;
   gstelement_class = (GstElementClass *) klass;
   gstbasesrc_class = GST_BASE_SRC_CLASS (klass);
-  gstpush_src_class = GST_PUSH_SRC_CLASS (klass);
 
   gobject_class->finalize = rsn_dvdsrc_finalize;
   gobject_class->set_property = rsn_dvdsrc_set_property;
@@ -228,7 +230,9 @@ rsn_dvdsrc_class_init (resinDvdSrcClass * klass)
       GST_DEBUG_FUNCPTR (rsn_dvdsrc_prepare_seek);
   gstbasesrc_class->do_seek = GST_DEBUG_FUNCPTR (rsn_dvdsrc_do_seek);
 
-  gstpush_src_class->create = GST_DEBUG_FUNCPTR (rsn_dvdsrc_create);
+  gstbasesrc_class->check_get_range =
+      GST_DEBUG_FUNCPTR (rsn_dvdsrc_check_get_range);
+  gstbasesrc_class->create = GST_DEBUG_FUNCPTR (rsn_dvdsrc_create);
 
   g_object_class_install_property (gobject_class, ARG_DEVICE,
       g_param_spec_string ("device", "Device", "DVD device location",
@@ -257,7 +261,7 @@ rsn_dvdsrc_init (resinDvdSrc * rsndvdsrc, resinDvdSrcClass * gclass)
   rsndvdsrc->branching = FALSE;
   rsndvdsrc->still_cond = g_cond_new ();
 
-  rsn_base_src_set_format (GST_BASE_SRC (rsndvdsrc), GST_FORMAT_TIME);
+  gst_base_src_set_format (GST_BASE_SRC (rsndvdsrc), GST_FORMAT_TIME);
 }
 
 static void
@@ -276,7 +280,7 @@ rsn_dvdsrc_finalize (GObject * object)
 }
 
 static gboolean
-rsn_dvdsrc_unlock (RsnBaseSrc * bsrc)
+rsn_dvdsrc_unlock (GstBaseSrc * bsrc)
 {
   resinDvdSrc *src = RESINDVDSRC (bsrc);
 
@@ -289,7 +293,7 @@ rsn_dvdsrc_unlock (RsnBaseSrc * bsrc)
 }
 
 static gboolean
-rsn_dvdsrc_unlock_stop (RsnBaseSrc * bsrc)
+rsn_dvdsrc_unlock_stop (GstBaseSrc * bsrc)
 {
   resinDvdSrc *src = RESINDVDSRC (bsrc);
 
@@ -351,7 +355,7 @@ rsn_dvdsrc_get_property (GObject * object, guint prop_id,
 }
 
 static gboolean
-rsn_dvdsrc_start (RsnBaseSrc * bsrc)
+rsn_dvdsrc_start (GstBaseSrc * bsrc)
 {
   resinDvdSrc *src = RESINDVDSRC (bsrc);
 
@@ -508,7 +512,7 @@ get_vts_attr (resinDvdSrc * src, gint n)
 }
 
 static gboolean
-rsn_dvdsrc_stop (RsnBaseSrc * bsrc)
+rsn_dvdsrc_stop (GstBaseSrc * bsrc)
 {
   resinDvdSrc *src = RESINDVDSRC (bsrc);
   gboolean ret = TRUE;
@@ -594,7 +598,6 @@ rsn_dvdsrc_do_still (resinDvdSrc * src, int duration)
   GstEvent *still_event;
   GstEvent *hl_event;
   gboolean cmds_changed;
-  GstStructure *s;
   GstEvent *seg_event;
   GstSegment *segment = &(GST_BASE_SRC (src)->segment);
 
@@ -607,12 +610,9 @@ rsn_dvdsrc_do_still (resinDvdSrc * src, int duration)
     else
       src->still_time_remaining = GST_SECOND * duration;
 
-    /* Send a close-segment event, and a dvd-still start
+    /* Send a close-segment event, and a still-frame start
      * event, then sleep */
-    s = gst_structure_new ("application/x-gst-dvd",
-        "event", G_TYPE_STRING, "dvd-still",
-        "still-state", G_TYPE_BOOLEAN, TRUE, NULL);
-    still_event = gst_event_new_custom (GST_EVENT_CUSTOM_DOWNSTREAM, s);
+    still_event = gst_video_event_new_still_frame (TRUE);
 
     gst_segment_set_last_stop (segment, GST_FORMAT_TIME, src->cur_end_ts);
 
@@ -731,10 +731,7 @@ rsn_dvdsrc_do_still (resinDvdSrc * src, int duration)
 
     /* Tell downstream the still is over.
      * We only do this if the still isn't interrupted: */
-    s = gst_structure_new ("application/x-gst-dvd",
-        "event", G_TYPE_STRING, "dvd-still",
-        "still-state", G_TYPE_BOOLEAN, FALSE, NULL);
-    still_event = gst_event_new_custom (GST_EVENT_CUSTOM_DOWNSTREAM, s);
+    still_event = gst_video_event_new_still_frame (FALSE);
 
     /* If the segment was too short in a timed still, it may need extending */
     if (segment->last_stop < segment->start + GST_SECOND * duration)
@@ -1069,9 +1066,8 @@ rsn_dvdsrc_step (resinDvdSrc * src, gboolean have_dvd_lock)
       break;
     }
     case DVDNAV_HIGHLIGHT:{
-      dvdnav_highlight_event_t *event = (dvdnav_highlight_event_t *) data;
       GST_DEBUG_OBJECT (src, "highlight change event, button %d",
-          event->buttonN);
+          ((dvdnav_highlight_event_t *) data)->buttonN);
       rsn_dvdsrc_update_highlight (src);
       break;
     }
@@ -1226,10 +1222,19 @@ rsn_dvdsrc_prepare_next_block (resinDvdSrc * src, gboolean have_dvd_lock)
   return ret;
 }
 
-static GstFlowReturn
-rsn_dvdsrc_create (RsnPushSrc * psrc, GstBuffer ** outbuf)
+static gboolean
+rsn_dvdsrc_check_get_range (GstBaseSrc * src)
 {
-  resinDvdSrc *src = RESINDVDSRC (psrc);
+  /* ResinDVD never operates in pull mode. There might be
+   * a reason to in the future though? */
+  return FALSE;
+}
+
+static GstFlowReturn
+rsn_dvdsrc_create (GstBaseSrc * bsrc, guint64 offset,
+    guint length, GstBuffer ** outbuf)
+{
+  resinDvdSrc *src = RESINDVDSRC (bsrc);
   GstSegment *segment = &(GST_BASE_SRC (src)->segment);
   GstFlowReturn ret;
   GstEvent *streams_event = NULL;
@@ -1299,19 +1304,30 @@ rsn_dvdsrc_create (RsnPushSrc * psrc, GstBuffer ** outbuf)
 
   if (src->need_segment) {
     /* Seamless segment update */
-    GstEvent *seek;
+    GstClockTime position = 0;
+
+    if (src->cur_position != GST_CLOCK_TIME_NONE)
+      position += src->cur_position;
+    if (src->cur_vobu_base_ts != GST_CLOCK_TIME_NONE)
+      position += src->cur_vobu_base_ts;
+
+    GST_DEBUG_OBJECT (src,
+        "Starting seamless segment update to %" GST_TIME_FORMAT " -> %"
+        GST_TIME_FORMAT " VOBU %" GST_TIME_FORMAT " position %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (src->cur_start_ts), GST_TIME_ARGS (src->cur_end_ts),
+        GST_TIME_ARGS (src->cur_vobu_base_ts), GST_TIME_ARGS (position));
+
+    gst_base_src_new_seamless_segment (GST_BASE_SRC (src),
+        src->cur_start_ts, -1, position);
 
-    seek = gst_event_new_seek (segment->rate, rsndvd_format,
-        GST_SEEK_FLAG_NONE, GST_SEEK_TYPE_NONE, -1, GST_SEEK_TYPE_NONE, -1);
-    gst_element_send_event (GST_ELEMENT (src), seek);
     src->need_segment = FALSE;
   }
 
-  g_mutex_lock (src->dvd_lock);
-
   if (src->cur_end_ts != GST_CLOCK_TIME_NONE)
     gst_segment_set_last_stop (segment, GST_FORMAT_TIME, src->cur_end_ts);
 
+  g_mutex_lock (src->dvd_lock);
+
   if (src->next_buf != NULL) {
     /* Now that we're in the new segment, we can enqueue any nav packet
      * correctly */
@@ -1562,15 +1578,11 @@ rsn_dvdsrc_handle_navigation_event (resinDvdSrc * src, GstEvent * event)
         gint title = 0;
         gint part = 0;
 
-        if (dvdnav_current_title_info (src->dvdnav, &title, &part) && title > 0
-            && part > 1) {
-          if (dvdnav_part_play (src->dvdnav, title, part - 1) ==
-              DVDNAV_STATUS_ERR)
+        if (dvdnav_current_title_info (src->dvdnav, &title, &part)) {
+          if (title > 0 && part > 1) {
             dvdnav_prev_pg_search (src->dvdnav);
-          nav_res = RSN_NAV_RESULT_BRANCH;
-        } else {
-          dvdnav_prev_pg_search (src->dvdnav);
-          nav_res = RSN_NAV_RESULT_BRANCH;
+            nav_res = RSN_NAV_RESULT_BRANCH;
+          }
         }
       } else if (g_str_equal (key, "period")) {
         dvdnav_next_pg_search (src->dvdnav);
@@ -1581,6 +1593,10 @@ rsn_dvdsrc_handle_navigation_event (resinDvdSrc * src, GstEvent * event)
       } else if (g_str_equal (key, "bracketright")) {
         nav_res =
             rsn_dvdsrc_do_command (src, GST_NAVIGATION_COMMAND_NEXT_ANGLE);
+      } else if (key && key[0] >= '1' && key[0] <= '8') {
+        gint new_stream = key[0] - '1';
+        GST_INFO_OBJECT (src, "Selecting audio stream %d", new_stream);
+        rsn_dvdsrc_prepare_audio_stream_event (src, new_stream, new_stream);
       }
       break;
     }
@@ -1814,6 +1830,7 @@ rsn_dvdsrc_prepare_streamsinfo_event (resinDvdSrc * src)
   audio_attr_t *a_attrs;
   subp_attr_t *s_attrs;
   gint n_audio, n_subp;
+  int8_t cur_audio;
   GstStructure *s;
   GstEvent *e;
   gint i;
@@ -1868,6 +1885,8 @@ rsn_dvdsrc_prepare_streamsinfo_event (resinDvdSrc * src)
       NULL);
 
   /* audio */
+  cur_audio = dvdnav_get_active_audio_stream (src->dvdnav);
+
   have_audio = FALSE;
   for (i = 0; i < n_audio; i++) {
     const audio_attr_t *a = a_attrs + i;
@@ -1881,18 +1900,23 @@ rsn_dvdsrc_prepare_streamsinfo_event (resinDvdSrc * src)
 
     GST_DEBUG_OBJECT (src, "mapped logical audio %d to MPEG substream %d",
         i, phys_id);
-
+    /* Force audio stream reselection in case format changed ... */
+    if (i == cur_audio) {
+      src->cur_audio_phys_stream = -1;
+      rsn_dvdsrc_prepare_audio_stream_event (src, i, phys_id);
+    }
 #if 0
-    /* FIXME: Only output A52 streams for now, until the decoder switching
-     * is ready */
+    /* Old test code: Only output A52 streams */
     if (a->audio_format != 0) {
       GST_DEBUG_OBJECT (src, "Ignoring non-A52 stream %d, format %d", i,
           (int) a->audio_format);
       continue;
     }
-#endif
     if (a->audio_format == 0)
       have_audio = TRUE;
+#else
+    have_audio = TRUE;
+#endif
 
     GST_DEBUG_OBJECT (src, "Audio stream %d is format %d, substream %d", i,
         (int) a->audio_format, phys_id);
@@ -1905,7 +1929,9 @@ rsn_dvdsrc_prepare_streamsinfo_event (resinDvdSrc * src)
     gst_structure_set (s, t, G_TYPE_INT, (int) a->audio_format, NULL);
     g_free (t);
 
-    if (a->lang_type) {
+    /* Check that the language code is flagged and at least somewhat valid
+     * before putting it in the output structure */
+    if (a->lang_type && a->lang_code > 0x100) {
       t = g_strdup_printf ("audio-%d-language", i);
       lang_code[0] = (a->lang_code >> 8) & 0xff;
       lang_code[1] = a->lang_code & 0xff;
@@ -1914,7 +1940,7 @@ rsn_dvdsrc_prepare_streamsinfo_event (resinDvdSrc * src)
 
       GST_DEBUG_OBJECT (src, "Audio stream %d is language %s", i, lang_code);
     } else
-      GST_DEBUG_OBJECT (src, "Audio stream %d - no language %s", i, lang_code);
+      GST_DEBUG_OBJECT (src, "Audio stream %d - no language", i);
   }
 
   if (have_audio == FALSE) {
@@ -1948,7 +1974,7 @@ rsn_dvdsrc_prepare_streamsinfo_event (resinDvdSrc * src)
     g_free (t);
 
     t = g_strdup_printf ("subpicture-%d-language", i);
-    if (u->type) {
+    if (u->type && u->lang_code > 0x100) {
       lang_code[0] = (u->lang_code >> 8) & 0xff;
       lang_code[1] = u->lang_code & 0xff;
       gst_structure_set (s, t, G_TYPE_STRING, lang_code, NULL);
@@ -2235,7 +2261,7 @@ rsn_dvdsrc_nav_clock_cb (GstClock * clock, GstClockTime time, GstClockID id,
   return TRUE;
 }
 
-/* Called with dvd_lock held */
+/* Called with dvd_lock held. NOTE: Releases dvd_lock briefly */
 static void
 rsn_dvdsrc_schedule_nav_cb (resinDvdSrc * src, RsnDvdPendingNav * next_nav)
 {
@@ -2266,8 +2292,10 @@ rsn_dvdsrc_schedule_nav_cb (resinDvdSrc * src, RsnDvdPendingNav * next_nav)
   GST_LOG_OBJECT (src, "Schedule nav pack for running TS %" GST_TIME_FORMAT,
       GST_TIME_ARGS (next_nav->running_ts));
 
+  g_mutex_unlock (src->dvd_lock);
   gst_clock_id_wait_async (src->nav_clock_id, rsn_dvdsrc_nav_clock_cb, src);
   gst_object_unref (clock);
+  g_mutex_lock (src->dvd_lock);
 }
 
 /* Called with dvd_lock held */
@@ -2293,7 +2321,7 @@ rsn_dvdsrc_check_nav_blocks (resinDvdSrc * src)
 }
 
 static gboolean
-rsn_dvdsrc_src_event (RsnBaseSrc * basesrc, GstEvent * event)
+rsn_dvdsrc_src_event (GstBaseSrc * basesrc, GstEvent * event)
 {
   resinDvdSrc *src = RESINDVDSRC (basesrc);
   gboolean res;
@@ -2369,7 +2397,7 @@ rsn_dvdsrc_change_state (GstElement * element, GstStateChange transition)
 }
 
 static gboolean
-rsn_dvdsrc_src_query (RsnBaseSrc * basesrc, GstQuery * query)
+rsn_dvdsrc_src_query (GstBaseSrc * basesrc, GstQuery * query)
 {
   resinDvdSrc *src = RESINDVDSRC (basesrc);
   gboolean res = FALSE;
@@ -2462,13 +2490,13 @@ rsn_dvdsrc_src_query (RsnBaseSrc * basesrc, GstQuery * query)
 }
 
 static gboolean
-rsn_dvdsrc_is_seekable (RsnBaseSrc * bsrc)
+rsn_dvdsrc_is_seekable (GstBaseSrc * bsrc)
 {
   return TRUE;
 }
 
 static gboolean
-rsn_dvdsrc_prepare_seek (RsnBaseSrc * bsrc, GstEvent * event,
+rsn_dvdsrc_prepare_seek (GstBaseSrc * bsrc, GstEvent * event,
     GstSegment * segment)
 {
   GstSeekType cur_type, stop_type;
@@ -2497,7 +2525,7 @@ rsn_dvdsrc_prepare_seek (RsnBaseSrc * bsrc, GstEvent * event,
     return TRUE;
   }
 
-  /* Let basesrc handle other formats for now. FIXME: Implement angle */
+  /* Let basesrc handle other formats */
   ret = GST_BASE_SRC_CLASS (parent_class)->prepare_seek_segment (bsrc,
       event, segment);
 
@@ -2557,6 +2585,9 @@ rsn_dvdsrc_get_sector_from_time_tmap (resinDvdSrc * src, GstClockTime ts)
 
   /* Get the time map */
   title_tmap = vts_tmapt->tmap + vts_ttn - 1;
+  if (title_tmap->tmu == 0)
+    return -1;
+
   entry = ts / (title_tmap->tmu * GST_SECOND);
   if (entry == 0)
     return 0;
@@ -2624,7 +2655,7 @@ rsn_dvdsrc_seek_to_time (resinDvdSrc * src, GstClockTime ts)
 }
 
 static gboolean
-rsn_dvdsrc_do_seek (RsnBaseSrc * bsrc, GstSegment * segment)
+rsn_dvdsrc_do_seek (GstBaseSrc * bsrc, GstSegment * segment)
 {
   resinDvdSrc *src = RESINDVDSRC (bsrc);
   gboolean ret = FALSE;
@@ -2634,10 +2665,10 @@ rsn_dvdsrc_do_seek (RsnBaseSrc * bsrc, GstSegment * segment)
      * everything up and flushing, we just need to step to the next
      * data block (below) so we know our new position */
     ret = TRUE;
+    /* HACK to make initial seek work: */
     src->first_seek = FALSE;
   } else {
-    /* FIXME: Handle other formats: Time, title, chapter, angle */
-    /* HACK to make initial seek work: */
+    /* Handle other formats: Time, title, chapter, angle */
     if (segment->format == GST_FORMAT_TIME) {
       g_mutex_lock (src->dvd_lock);
       src->discont = TRUE;
@@ -2718,8 +2749,12 @@ rsn_dvdsrc_do_seek (RsnBaseSrc * bsrc, GstSegment * segment)
 
     GST_LOG_OBJECT (src, "Entering prepare_next_block after seek."
         " Flushing = %d", src->flushing_seek);
-    if (rsn_dvdsrc_prepare_next_block (src, FALSE) != GST_FLOW_OK)
-      goto fail;
+    while (src->cur_start_ts == GST_CLOCK_TIME_NONE) {
+      if (rsn_dvdsrc_prepare_next_block (src, FALSE) != GST_FLOW_OK)
+        goto fail;
+      if (src->cur_start_ts == GST_CLOCK_TIME_NONE)
+        gst_buffer_replace (&src->next_buf, NULL);
+    }
     GST_LOG_OBJECT (src, "prepare_next_block after seek done");
 
     segment->format = GST_FORMAT_TIME;
