@@ -47,10 +47,7 @@
 #include <unistd.h>
 
 #include <gst/gst.h>
-#include <glade/glade-xml.h>
 #include <gtk/gtk.h>
-#include <gst/interfaces/xoverlay.h>
-#include <gdk/gdkx.h>
 
 /*
  * Global constants
@@ -86,8 +83,7 @@ typedef enum _AppOptions {
 /* gstreamer related functions */
 
 static void me_gst_cleanup_elements ();
-static int
-me_gst_setup_view_pipeline (const gchar * filename, GdkWindow * window);
+static int me_gst_setup_view_pipeline (const gchar * filename);
 static int
 me_gst_setup_capture_pipeline (const gchar * src_file, const gchar * dest_file,
     gint * encode_status, gboolean use_v4l2);
@@ -115,11 +111,14 @@ GstElement *gst_video_sink = NULL;
 GstElement *gst_file_sink = NULL;
 GstElement *gst_pipeline = NULL;
 
+GdkPixbuf *last_pixbuf = NULL;  /* image as pixbuf at original size */
+GdkPixbuf *draw_pixbuf = NULL;  /* pixbuf resized for drawing       */
+
 AppOptions app_options = APP_OPT_ALL;
 
 GstTagList *tag_list = NULL;
 
-GladeXML *ui_glade_xml = NULL;
+GtkBuilder *builder = NULL;
 GtkWidget *ui_main_window = NULL;
 GtkWidget *ui_drawing = NULL;
 GtkWidget *ui_tree = NULL;
@@ -136,6 +135,30 @@ GString *filename = NULL;
  * Helper functions
  */
 
+static void
+dump_tag_buffer(const char *tag, guint8 * buf, guint32 size)
+{
+  guint32 i;
+  printf("\nDumping %s (size = %u)\n\n", tag, size);
+
+  for(i=0; i<size; ++i) {
+
+    if (i % 16 == 0)
+      printf("%04x:%04x | ", i >> 16, i & 0xFFFF);
+
+    printf("%02x", buf[i]);
+
+    if (i % 16 != 15)
+      printf(" ");
+    else
+      printf("\n");
+
+  }
+
+  printf("\n\n");
+
+}
+
 static void
 insert_tag_on_tree (const GstTagList * list, const gchar * tag,
     gpointer user_data)
@@ -150,6 +173,13 @@ insert_tag_on_tree (const GstTagList * list, const gchar * tag,
   if (gst_tag_get_type (tag) == G_TYPE_STRING) {
     if (!gst_tag_list_get_string_index (list, tag, 0, &str))
       g_assert_not_reached ();
+  } else if ( gst_tag_get_type (tag) == GST_TYPE_BUFFER ) {
+    const GValue *val = NULL;
+    GstBuffer *buf;
+    val = gst_tag_list_get_value_index (list, tag, 0);
+    buf = gst_value_get_buffer (val);
+    dump_tag_buffer(tag, GST_BUFFER_DATA(buf), GST_BUFFER_SIZE(buf));
+    str = g_strdup("It has been printed to stdout");
   } else {
     str = g_strdup_value_contents (gst_tag_list_get_value_index (list, tag, 0));
   }
@@ -215,7 +245,8 @@ change_tag_list (GstTagList ** list, const gchar * tag, const gchar * value)
       }
         break;
       default:
-        g_printerr ("Tags of type '%s' are not supported yet.\n",
+        g_printerr ("Tags of type '%s' are not supported yet for editing"
+                    " by this application.\n",
             g_type_name (type));
         break;
     }
@@ -227,39 +258,88 @@ done:
 
 }
 
-
-/*
- * UI handling functions (mapped by glade)
- */
-
-gboolean
-on_windowMain_configure_event (GtkWidget * widget, GdkEventConfigure * event,
-    gpointer user_data)
+static void
+update_draw_pixbuf (guint max_width, guint max_height)
 {
-  GstXOverlay *xoverlay;
-
-  if (gst_video_sink) {
-
-    xoverlay = GST_X_OVERLAY (gst_video_sink);
-
-    if (xoverlay != NULL && GST_IS_X_OVERLAY (xoverlay)) {
-      gst_x_overlay_expose (xoverlay);
-    }
+  gdouble wratio, hratio;
+  gint w = 0, h = 0;
 
+  if (last_pixbuf) {
+    w = gdk_pixbuf_get_width (last_pixbuf);
+    h = gdk_pixbuf_get_height (last_pixbuf);
   }
 
-  return FALSE;
+  g_print ("Allocation: %dx%d, pixbuf: %dx%d\n", max_width, max_height, w, h);
+
+  if (last_pixbuf == NULL)
+    return;
+
+  g_return_if_fail (max_width > 0 && max_height > 0);
+  wratio = w * 1.0 / max_width * 1.0;
+  hratio = h * 1.0 / max_height;
+  g_print ("ratios = %.2f / %.2f\n", wratio, hratio);
+
+  if (hratio > wratio) {
+    w = (gint) (w * 1.0 / hratio);
+    h = (gint) (h * 1.0 / hratio);
+  } else {
+    w = (gint) (w * 1.0 / wratio);
+    h = (gint) (h * 1.0 / wratio);
+  }
+
+  if (draw_pixbuf != NULL && gdk_pixbuf_get_width (draw_pixbuf) == w &&
+      gdk_pixbuf_get_height (last_pixbuf) == h) {
+    return; /* nothing to do */
+  }
+
+  g_print ("drawing pixbuf at %dx%d\n", w, h);
+
+  if (draw_pixbuf)
+    g_object_unref (draw_pixbuf);
+
+  draw_pixbuf =
+      gdk_pixbuf_scale_simple (last_pixbuf, w, h, GDK_INTERP_BILINEAR);
+  
 }
 
+static void
+ui_drawing_size_allocate_cb (GtkWidget * drawing_area,
+    GtkAllocation * allocation, gpointer data)
+{
+  update_draw_pixbuf (allocation->width, allocation->height);
+}
+
+/*
+ * UI handling functions (mapped by GtkBuilder)
+ */
+
 gboolean
 on_drawingMain_expose_event (GtkWidget * widget, GdkEventExpose * event,
     gpointer data)
 {
-  if (gst_video_sink) {
-    gst_x_overlay_set_xwindow_id (GST_X_OVERLAY (gst_video_sink),
-        GDK_WINDOW_XWINDOW (widget->window));
-  }
-  return FALSE;
+  GtkAllocation a = widget->allocation;
+  gint w, h, x, y;
+
+  if (draw_pixbuf == NULL)
+    return FALSE;
+
+  w = gdk_pixbuf_get_width (draw_pixbuf);
+  h = gdk_pixbuf_get_height (draw_pixbuf);
+
+  /* center image */
+  x = (a.width - w) / 2;
+  y = (a.height - h) / 2;
+
+  /* sanity check, shouldn't happen */
+  if (x < 0)
+    x = 0;
+  if (y < 0)
+    y = 0;
+
+  gdk_draw_pixbuf (GDK_DRAWABLE (widget->window), widget->style->black_gc,
+      draw_pixbuf, 0, 0, x, y, w, h, GDK_RGB_DITHER_NONE, 0, 0);
+
+  return TRUE; /* handled expose event */
 }
 
 void
@@ -389,7 +469,7 @@ on_buttonSaveFile_clicked (GtkButton * button, gpointer user_data)
       tag_list = NULL;
     }
 
-    me_gst_setup_view_pipeline (filename->str, ui_drawing->window);
+    me_gst_setup_view_pipeline (filename->str);
 
     gst_element_set_state (gst_pipeline, GST_STATE_PLAYING);
     gst_element_get_state (gst_pipeline, NULL, NULL, GST_CLOCK_TIME_NONE);
@@ -659,94 +739,40 @@ ui_refresh ()
     gtk_window_set_title (GTK_WINDOW (ui_main_window), filename->str);
 }
 
-static int
-ui_connect_signals()
-{
-
-  glade_xml_signal_connect(ui_glade_xml, "on_checkbuttonCaptureV4l2_toggled",
-      (GCallback)on_checkbuttonCaptureV4l2_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_checkbuttonCaptureTest_toggled",
-      (GCallback)on_checkbuttonCaptureTest_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml,
-      "on_checkbuttonOptionsDemuxExif_toggled",
-      (GCallback) on_checkbuttonOptionsDemuxExif_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml,
-      "on_checkbuttonOptionsDemuxIptc_toggled",
-      (GCallback) on_checkbuttonOptionsDemuxIptc_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml,
-      "on_checkbuttonOptionsDemuxXmp_toggled",
-      (GCallback) on_checkbuttonOptionsDemuxXmp_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml,
-      "on_checkbuttonOptionsMuxExif_toggled",
-      (GCallback) on_checkbuttonOptionsMuxExif_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml,
-      "on_checkbuttonOptionsMuxIptc_toggled",
-      (GCallback) on_checkbuttonOptionsMuxIptc_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml,
-      "on_checkbuttonOptionsMuxXmp_toggled",
-      (GCallback) on_checkbuttonOptionsMuxXmp_toggled);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_buttonSaveFile_clicked",
-      (GCallback)on_buttonSaveFile_clicked);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_windowMain_delete_event",
-      (GCallback)on_windowMain_delete_event);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_drawingMain_expose_event",
-      (GCallback)on_drawingMain_expose_event);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_windowMain_configure_event",
-      (GCallback)on_windowMain_configure_event);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_buttonInsert_clicked",
-      (GCallback)on_buttonInsert_clicked);
-
-  glade_xml_signal_connect(ui_glade_xml, "on_buttonOpenFile_clicked",
-      (GCallback)on_buttonOpenFile_clicked);
-
-  return 0;
-
-}
-
 static int
 ui_create ()
 {
+  GError *error = NULL;
   int ret = 0;
 
-  ui_glade_xml = glade_xml_new ("metadata_editor.glade", NULL, NULL);
-
-  if (!ui_glade_xml) {
-    fprintf (stderr, "glade_xml_new failed\n");
+  builder = gtk_builder_new ();
+  if (!gtk_builder_add_from_file (builder, "metadata_editor.ui", &error))
+  {
+    g_warning ("Couldn't load builder file: %s", error->message);
+    g_error_free (error);
     ret = -101;
     goto done;
   }
 
-  ui_main_window = glade_xml_get_widget (ui_glade_xml, "windowMain");
+  ui_main_window = GTK_WIDGET (gtk_builder_get_object (builder, "windowMain"));
 
-  ui_drawing = glade_xml_get_widget (ui_glade_xml, "drawingMain");
+  ui_drawing = GTK_WIDGET (gtk_builder_get_object (builder, "drawingMain"));
 
-  ui_tree = glade_xml_get_widget (ui_glade_xml, "treeMain");
+  ui_tree = GTK_WIDGET (gtk_builder_get_object (builder, "treeMain"));
 
   ui_entry_insert_tag =
-      GTK_ENTRY (glade_xml_get_widget (ui_glade_xml, "entryTag"));
+          GTK_ENTRY (gtk_builder_get_object (builder, "entryTag"));
 
   ui_entry_insert_value =
-      GTK_ENTRY (glade_xml_get_widget (ui_glade_xml, "entryValue"));
+          GTK_ENTRY (gtk_builder_get_object (builder, "entryValue"));
 
   ui_chk_bnt_capture_v4l2 =
-      GTK_TOGGLE_BUTTON (glade_xml_get_widget (ui_glade_xml,
-          "checkbuttonCaptureV4l2"));
+          GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder,
+                                                     "checkbuttonCaptureV4l2"));
 
   ui_chk_bnt_capture_test =
-      GTK_TOGGLE_BUTTON (glade_xml_get_widget (ui_glade_xml,
-          "checkbuttonCaptureTest"));
+          GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder,
+                                                     "checkbuttonCaptureTest"));
 
   if (!(ui_main_window && ui_drawing && ui_tree
           && ui_entry_insert_tag && ui_entry_insert_value
@@ -756,7 +782,10 @@ ui_create ()
     goto done;
   }
 
-  ui_connect_signals();
+  g_signal_connect_after (ui_drawing, "size-allocate",
+      G_CALLBACK (ui_drawing_size_allocate_cb), NULL);
+
+  gtk_builder_connect_signals (builder, NULL);
 
   ui_setup_tree_view (GTK_TREE_VIEW (ui_tree));
 
@@ -851,8 +880,8 @@ me_gst_bus_callback_view (GstBus * bus, GstMessage * message, gpointer data)
           if (ntl) {
             gst_tag_list_free (tag_list);
             tag_list = ntl;
-            gst_tag_list_free (tl);
           }
+          gst_tag_list_free (tl);
         }
       }
       /* remove whole chunk tags */
@@ -866,6 +895,37 @@ me_gst_bus_callback_view (GstBus * bus, GstMessage * message, gpointer data)
         gst_tag_list_foreach (tag_list, insert_tag_on_tree, ui_tree);
       }
       break;
+    case GST_MESSAGE_ELEMENT: {
+      const GValue *val;
+
+      /* only interested in element messages from our gdkpixbufsink */
+      if (message->src != GST_OBJECT_CAST (gst_video_sink))
+        break;
+
+      /* only interested in the first image (not any smaller previews) */
+      if (gst_structure_has_name (message->structure, "pixbuf"))
+        break;
+
+      if (!gst_structure_has_name (message->structure, "preroll-pixbuf"))
+        break;
+
+      val = gst_structure_get_value (message->structure, "pixbuf");
+      g_return_val_if_fail (val != NULL, TRUE);
+
+      if (last_pixbuf)
+        g_object_unref (last_pixbuf);
+
+      last_pixbuf = GDK_PIXBUF (g_value_dup_object (val));
+
+      g_print ("Got image pixbuf: %dx%d\n", gdk_pixbuf_get_width (last_pixbuf),
+          gdk_pixbuf_get_height (last_pixbuf));
+
+      update_draw_pixbuf (GTK_WIDGET (ui_drawing)->allocation.width,
+          GTK_WIDGET (ui_drawing)->allocation.height);
+
+      gtk_widget_queue_draw (ui_drawing);
+      break;
+    }
     default:
       /* unhandled message */
       break;
@@ -1139,7 +1199,7 @@ done:
 }
 
 static int
-me_gst_setup_view_pipeline (const gchar * filename, GdkWindow * window)
+me_gst_setup_view_pipeline (const gchar * filename)
 {
   int ret = 0;
   GstBus *bus = NULL;
@@ -1157,8 +1217,18 @@ me_gst_setup_view_pipeline (const gchar * filename, GdkWindow * window)
     gst_image_dec = gst_element_factory_make ("jpegdec", NULL);
   gst_video_scale = gst_element_factory_make ("videoscale", NULL);
   gst_video_convert = gst_element_factory_make ("ffmpegcolorspace", NULL);
-  gst_video_sink = gst_element_factory_make ("ximagesink", NULL);
+  gst_video_sink = gst_element_factory_make ("gdkpixbufsink", NULL);
 
+  if (gst_video_sink == NULL) {
+    if (!gst_default_registry_check_feature_version ("gdkpixbufdec", 0, 10, 0))
+      g_warning ("Could not create 'gdkpixbufsink' element");
+    else {
+      g_warning ("Could not create 'gdkpixbufsink' element. "
+          "(May be your gst-plugins-good is too old?)");
+    ret = -400;
+    }
+    goto done;
+  }
   if (!(gst_source && gst_metadata_demux && gst_image_dec && gst_video_scale
           && gst_video_convert && gst_video_sink)) {
     fprintf (stderr, "An element couldn't be created for viewing\n");
@@ -1178,7 +1248,6 @@ me_gst_setup_view_pipeline (const gchar * filename, GdkWindow * window)
   /* set elements's properties */
   g_object_set (gst_source, "location", filename, NULL);
   g_object_set (gst_metadata_demux, "parse-only", TRUE, NULL);
-  g_object_set (gst_video_sink, "force-aspect-ratio", TRUE, NULL);
 
 
   /* adding and linking elements */
@@ -1193,9 +1262,10 @@ me_gst_setup_view_pipeline (const gchar * filename, GdkWindow * window)
       gst_video_convert = NULL;
   gst_object_ref (gst_video_sink);
 
-  if (window)
-    gst_x_overlay_set_xwindow_id (GST_X_OVERLAY (gst_video_sink),
-        GDK_WINDOW_XWINDOW (window));
+  if (last_pixbuf) {
+    g_object_unref (last_pixbuf);
+    last_pixbuf = NULL;
+  }
 
   if (!linked) {
     fprintf (stderr, "Elements couldn't be linked\n");
@@ -1226,7 +1296,7 @@ process_file() {
   }
 
   /* create pipeline */
-  me_gst_setup_view_pipeline (filename->str, ui_drawing->window);
+  me_gst_setup_view_pipeline (filename->str);
 
   gst_element_set_state (gst_pipeline, GST_STATE_PLAYING);
 
